"""
User inputs can be provided by command line.
This package contains the logic to parse the command line arguments for internal
components based on the configuration items. but also the global arguments that are
hard coded.

The help method must be adapted to the action to be carried out.
Examples:
    `python weblodge.py -h` must print the `help` message of **WebLodge**.
    `python weblodge.py build -h` must print the `help` message for the `build` action.
"""
from .parameters import Parser, ConfigTrigger, ConfigIsDefined, ConfigIsNotDefined



"""
Internal utility functions for the web app.
"""
import os
import time
import logging

from dotenv import dotenv_values

from weblodge._azure import AzureWebApp, AzureService


logger = logging.getLogger('weblodge')


def get_webapp(azure_service: AzureService, subdomain: str) -> AzureWebApp:
    """
    Return a Azure Web App based on the subdomain.
    The Web App may not exists.
    """
    resource_group = azure_service.resource_groups(subdomain)
    keyvault = azure_service.keyvaults(subdomain, resource_group)
    return azure_service.web_apps(
        subdomain,
        resource_group=resource_group,
        app_service=azure_service.app_services(subdomain, resource_group),
        keyvault=keyvault
    )


def set_webapp_env_var(webapp: AzureWebApp, env_file: str, env_update_waiting_time: int) -> bool:
    """
    Udpate a Web App environment variable.
    Return True if the environment variable has been updated, False otherwise.
    """
    if os.path.exists(env_file):
        logger.info(f"Updating the environment variable with '{env_file}'...")
        env = dotenv_values(env_file)
        webapp.update_environment(env)
        logger.info('Environment variable updated.')
        logger.info('Waiting the application to restart...')
        time.sleep(env_update_waiting_time)
        return

    logger.info('No environment file found.')



# pylint: disable=consider-using-from-import

"""
Entry point for the CLI.

This module is the entry point for the CLI. It parses the command line arguments
and calls the appropriate functions.
"""
import sys
import logging
from typing import Dict
from pathlib import Path
from collections import defaultdict


import weblodge.state as state
from weblodge._azure import Service
from weblodge.parameters import Parser, ConfigIsNotDefined, ConfigIsDefined, ConfigTrigger
from weblodge.web_app import WebApp, NoMoreFreeApplicationAvailable, CanNotFindTierLocation, InvalidTier

from .args import get_cli_args, CLI_NAME


# Define the logger for internal usage.
# But the CLI will use print for user interraction.
logger = logging.getLogger('weblodge')
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())


# pylint: disable=missing-function-docstring
def main(return_web_app=False):
    success = False
    parameters = Parser()
    action, config_file = get_cli_args()
    web_app = WebApp(parameters.load, azure_service=Service())

    try:
        config = state.load(config_file)
        if action == 'build':
            success, config = web_app.build(config)
        elif action == 'clean':
            success = clean(parameters, web_app)
        elif action == 'deploy':
            success, config = deploy(config, web_app, parameters)
        elif action == 'delete':
            success = delete(config, web_app, parameters)
        elif action == 'github':
            success, config = github(config, web_app, config_file)
        elif action == 'list':
            list_(parameters.load, web_app)
            success = True
        elif action == 'logs':
            print('Logs will be stream, execute CTRL+C to stop the application.', flush=True)
            web_app.print_logs(config)
        elif action == 'app-tiers':
            success = list_app_tiers(config, web_app)
    except Exception as exception: # pylint: disable=broad-exception-caught
        print('Command failed with the following error:', exception, file=sys.stderr, flush=True)

    if success:
        state.dump(config_file, config)
        if return_web_app:
            return web_app
        return 0

    sys.exit(1)


def deploy(config: Dict[str, str], web_app: WebApp, parameters: Parser):
    """
    Deploy the application.
    """
    # The application can be built before being deployed.
    def _build(config):
        success, config = web_app.build(config)

        if not success:
            logger.critical('Deployment failed.')
            sys.exit(1)
        return config

    build_too = ConfigIsDefined(
        name='build',
        description='Build then deploy the application. Parameters are the same as for the `build` command.',
        attending_value=False,
        trigger=_build
    )

    parameters.trigger_once(build_too)
    try:
        success, config, tier = web_app.deploy(config)
    except NoMoreFreeApplicationAvailable as free_app_name:
        print(
            'Can not create the infrastrucutre. Azure support only one Free application by location.',
            'Please, change the deployment location or the application tier.',
            f"The already existing free application is this location is '{free_app_name}'.",
            file=sys.stderr,
            flush=True
        )
        return False, config
    except InvalidTier as _invalid_tier:
        print(
            'Invalid tier. Please, choose a tier from the following list:',
            file=sys.stderr,
            flush=True
        )
        list_app_tiers(config, web_app)
        return False, config

    print(f'Estimated cost: ${tier.price_by_hour * 730:.2f}/month')

    if success:
        print(f"The application will soon be available at: {web_app.url()}", flush=True)
    else:
        print(
            'The application may not be deployed, but the infrastructure may be' \
            f' partially created. You can delete it by running: {CLI_NAME} delete',
            sys=sys.stderr,
            flush=True
        )
    return success, config


def delete(config: Dict[str, str], web_app: WebApp, parameters: Parser):
    """
    Delete the application.
    """
    prompt = ConfigIsNotDefined(
        name='yes',
        description='Delete without user input.',
        trigger=_validation_before_deletion,
        attending_value=False
    )

    parameters.trigger_once(prompt)
    res, _ = web_app.delete(config)
    return res


def clean(parameters: Parser, web_app: WebApp):
    """
    Iterate over all resources and ask the user if he want to delete them.
    The parameter 'yes' is too risquy to be accepted and will be ignored.
    """
    prompt = ConfigTrigger(
        name='yes',
        description='The user input will always be asked',
        trigger=_validation_before_deletion,
        attending_value=False
    )

    for _wa in web_app.all(parameters.load):
        try:
            parameters.trigger_once(prompt)
            _wa.delete()
        except SystemExit:
            # User aborted the deletion.
            continue

    return True


def github(config: Dict[str, str], web_app: WebApp, config_file: str):
    """
    Create a GitHub application.
    """
    workflow_folder: str = '.github/workflows'
    workflow_file: str = '.github/workflows/weblodge.yml'

    config, workflow = web_app.github(config)

    if workflow:
        # Create the GitHub workflow folder if not exists.
        Path(workflow_folder).mkdir(parents=True, exist_ok=True)
        # Create the GitHub workflow file.
        Path(workflow_file).write_text(workflow.content, encoding='utf-8')

        # pylint: disable=line-too-long
        print(f'''Please, add the following secrets to your GitHub repository:
  - AZURE_CLIENT_ID: {workflow.client_id}
  - AZURE_TENANT_ID: {workflow.tenant_id}
  - AZURE_SUBSCRIPTION_ID: {workflow.subscription_id}
Then, commit and push the following files:
  - {workflow_file}
  - {config_file}
More information: https://docs.github.com/en/actions/security-guides/encrypted-secrets#creating-encrypted-secrets-for-a-repository''')

    return True, config


def list_app_tiers(config, web_app) -> bool:
    """
    Show to the user the available tiers.
    """
    try:
        # Retrieve the tiers.
        tiers = web_app.tiers(config)
    except CanNotFindTierLocation:
        print('Can not find any tier for the provided location.', file=sys.stderr, flush=True)
        print('Please, check the location and try again.', file=sys.stderr, flush=True)
        return False

    print('Warning: There is no guarantee of the estimated price.')

    # Group the tiers by description to print them by blocks.
    tiers_by_description = defaultdict(list)
    for tier in tiers:
        tiers_by_description[tier.description].append(tier)
        # Sort by cores to have a nice display.
        tiers_by_description[tier.description].sort(key=lambda t: t.cores)

    # Print the tiers.
    for description, tiers in tiers_by_description.items():
        # If the list is not empty, print the information.
        if tiers:
            print(f'\nTier description: {description}')
            print(f'Tier location: {tiers[0].location}')
            print(' Name |    Price    | Cores |   RAM   | Storage')
            print('-----------------------------------------------')
            for tier in tiers:
                print(f'{tier.name:>5} |  ${tier.price_by_hour:.2f}/hour |    {tier.cores:>2} | {tier.ram:>4} GB |  {str(tier.disk) + " GB":>6}')  # pylint: disable=line-too-long

    return True

def list_(parameter_loader, web_app: WebApp):
    """
    Print all the deployed applications.
    Warm the user if some infrastructure is not used.
    """
    unused_apps = []
    no_application_deployed = True

    for _wapp in web_app.all(parameter_loader):
        no_application_deployed = False
        if _wapp.exists():
            print(f"Application: {_wapp.url()}")
        else:
            unused_apps.append(_wapp.name)

    if no_application_deployed:
        print('No application deployed.')
        return

    if unused_apps:
        print('We found the following infrastructure without application deployed. This can be costly.')
        for name in unused_apps:
            print(f"Application '{name}', can be deleted by running: `{CLI_NAME} delete --subdomain {name}`")


def _validation_before_deletion(config):
    if input(f"Do you want to delete the application '{config['subdomain']}' (yes/no.)? ") != 'yes':
        print('Aborting.')
        sys.exit(0)
    return config



"""
Retrieve parameters of the application.

Currently, the parameters are retrieved from the command line.
"""
import argparse

from dataclasses import dataclass
from typing import Callable, Dict, List

from weblodge.config import Item as ConfigItem


@dataclass(frozen=True, eq=False)
class ConfigTrigger(ConfigItem):
    """
    Configuration Item with a function to trigger.
    Call the trigger whatever the config is defined or not.
    """
    # Function to call.
    trigger: Callable[[Dict[str, str]], Dict[str, str]] = None


@dataclass(frozen=True, eq=False)
class ConfigIsDefined(ConfigTrigger):
    """
    Call the trigger if the user provides the parameter.
    """


@dataclass(frozen=True, eq=False)
class ConfigIsNotDefined(ConfigTrigger):
    """
    Call the trigger if the user does not provide the parameter.
    """


class Parser:
    """
    Parse the command line arguments.

    Function(s) can be triggered if the user provides the right arguments.
    They will run only onces and their config will not be added to the global config.
    Example:
    ```
    def _build(config):
        print('Build the application.')
        return config
    parser = Parser()
    parser.trigger_once(
        ConfigTrigger(name='build', description='Build the application.', trigger=_build)
    )
    # If the user runs: `python weblodge.py build -h` then the help message will contain
    # the `build` option.
    # If the user runs: `python weblodge.py build` then the `_build` function will be called
    # but the config will not contains the 'build' item.
    parser.load(config)
    ```
    """
    def __init__(self) -> None:
        self._triggers: List[ConfigTrigger] = []

    # pylint: disable=too-many-branches
    def load(self, fields: List[ConfigItem], existing_parameters: Dict[str, str] = None) -> Dict[str, str]:
        """
        Load the configuration from the parser.
        Override the current config with the new values.
        """
        # The user can supply the same argument several times.
        # The ArgumentParser raise in this case, but that's not what we want.
        # If the user provides the same argument several times, we'll save the first one.
        argument_provided = set()
        existing_parameters = existing_parameters or {}

        # Create and configure a parser for the fields.
        parser = argparse.ArgumentParser()
        for field in fields + self._triggers:
            # Do not register twice the same argument.
            if field.name in argument_provided:
                continue
            argument_provided.add(field.name)

            argument = {
                'help': field.description,
            }

            default_value = existing_parameters.get(field.name, field.default)

            if not field.attending_value:
                argument['action'] = 'store_true'
            else:
                argument = {
                    **argument,
                    'type': str,
                    'required': default_value is None,
                    'default': default_value,
                }
                if field.values_allowed:
                    argument['choices'] = field.values_allowed

            parser.add_argument(
                f'--{_to_display(field.name)}',
                **argument
            )
        args_parsed, _ = parser.parse_known_args()
        new_config = vars(args_parsed)

        # Expected configuration must not contains configuration used for intermediate
        # operations.
        _complete_config = {**existing_parameters, **new_config}
        config = {k: v for k, v in _complete_config.items() if k not in self._triggers}

        # Trigger meta configs functions.
        while self._triggers:
            # Trigger are removes from the list as they are called
            # to avoid infinite recursion.
            _config = self._triggers.pop()
            if _config.attending_value:
                # If a value is expected, check if the user provided it.
                is_defined = _config.name in new_config
            else:
                # If no value is expected, the attending value will create
                # fill the config with False.
                is_defined = new_config[_config.name]

            if isinstance(_config, ConfigIsDefined):
                if is_defined:
                    config = _config.trigger(config)
            elif isinstance(_config, ConfigIsNotDefined):
                if not is_defined:
                    config = _config.trigger(config)
            elif isinstance(_config, ConfigTrigger):
                config = _config.trigger(config)

        return config

    def trigger_once(self, trigger: ConfigTrigger) -> 'Parser':
        """
        Add configuration trigger.
        Configuration defined will not be added to the config.
        This trigger will be analyse only once then remove from triggers.
        """
        self._triggers.append(trigger)


def _to_display(name: str) -> str:
    return name.replace('_', '-')



"""
Allow to retrieve all available tiers for a given location.
"""
from dataclasses import dataclass
import logging
from typing import List, Union

from weblodge.config import Item as ConfigItem
from weblodge._azure import AzureService, InvalidLocation

from .exceptions import CanNotFindTierLocation


logger = logging.getLogger('weblodge')


@dataclass
class WebAppTier:
    """
    Information on the WebApp hardware and price.
    """
    # Technical name of the tier.
    name: str

    # Name of the location where the tier is available.
    location: str

    # Price per hour of the tier.
    price_by_hour: float

    # Human description of the tier.
    description: str

    # Number of Cores.
    # It is a string when the SKU is free.
    cores: Union[int, str]

    # RAM in GB.
    ram: int

    # Disk size in GB.
    disk: int


class TiersConfig:
    """
    Tiers configuration.

    Tier or "SKU" on Azure is the name of the plan that will be used to host the application.
    It come with a price and hardware capabilities that may be asked by the user.
    """
    items = [
        ConfigItem(
            name='location',
            description='The physical location of tiers.',
            default='northeurope'
        ),
    ]

    # pylint: disable=too-many-arguments
    def __init__(
            self,
            location,
            *_args,
            **_kwargs
        ):
        # Tiers location.
        self.location = location


def tiers(azure_service: AzureService, config: TiersConfig) -> List[WebAppTier]:
    """
    Return all available tiers.
    """
    try:
        return [
            WebAppTier(
                name=s.name,
                location=s.location,
                price_by_hour=s.price_by_hour,
                description=s.description,
                cores=s.cores,
                ram=s.ram,
                disk=s.disk,
            )
            for s in azure_service.app_services.skus(config.location)
        ]
    except InvalidLocation:
        raise CanNotFindTierLocation(f"Can not find any tier for the location '{config.location}'.") from InvalidLocation  # pylint: disable=line-too-long



# pylint: disable=missing-module-docstring
import json


def dump(name_or_fd, config):
    """
    Dump the config to a file or file descriptor.
    """
    if isinstance(name_or_fd, str):
        with open(name_or_fd, 'w', encoding='utf-8') as file_descriptor:
            json.dump(config, file_descriptor, indent=4)
    else:
        json.dump(config, name_or_fd, indent=4)


def load(name_or_fd):
    """
    Load the config from a file or file descriptor.
    """
    try:
        if isinstance(name_or_fd, str):
            with open(name_or_fd, 'r', encoding='utf-8') as file_descriptor:
                return json.load(file_descriptor)
        else:
            return json.load(name_or_fd)
    except FileNotFoundError:
        return {}



"""
Internal Azure interface.

This package is for internal use only and must not be use from a third package.
"""
from .service import Service
from .exceptions import InvalidLocation
from .interfaces import AzureService, AzureAppServiceSku, \
    AzureResourceGroup, AzureWebApp, AzureAppService, \
    AzureLogLevel, MicrosoftEntraApplication, MicrosoftEntra, \
    AzureKeyVault, AzureKeyVaultSecret



"""
Web App facade.

Wrapp all actions related to the Azure Web App.
"""
import logging

from typing import Callable, Iterable, List, Dict, Optional, Tuple

from weblodge._azure import AzureService, AzureWebApp
from weblodge.config import Item as ConfigItem

from ._all import _all as _all_az_web_app
from .build import BuildConfig, build as _build
from .delete import DeleteConfig, delete as _delete
from .deploy import DeploymentConfig, deploy as _deploy
from .exceptions import RequirementsFileNotFound, EntryPointFileNotFound, FlaskAppNotFound, InvalidTier
from .logs import LogsConfig, logs as _logs
from .github import GitHubConfig, github, GitHubWorkflow
from .tiers import TiersConfig, tiers as _tiers, WebAppTier


logger = logging.getLogger('weblodge')


class WebApp:
    """
    Represent a Flask Web Application deploy on Azure.
    """
    def __init__(
        self,
        config_loader: Callable[[List[ConfigItem]], Dict[str, str]],
        azure_service: AzureService,
        web_app: AzureWebApp = None
    ):
        self.config_loader = config_loader
        self._web_app = web_app
        self.azure_service = azure_service

    @property
    def name(self) -> Optional[str]:
        """
        Return the name of the WebApp if exists, None otherwise.
        """
        return self._web_app.name if self._web_app else None

    def build(self, config: Dict[str, str]) -> Tuple[bool, Dict[str, str]]:
        """
        Build the application.
        """
        config = self.config_loader(BuildConfig.items, config)
        build_config = BuildConfig(**config)

        logger.info('Building...')
        try:
            _build(build_config)
        except RequirementsFileNotFound:
            logger.critical(f"Requirements file '{build_config.requirements}' not found.")
            logger.critical('Build failed.')
            return False, build_config
        except EntryPointFileNotFound:
            logger.critical(f"Entry point file '{build_config.entry_point}' not found.")
            logger.critical('Build failed.')
            return False, build_config
        except FlaskAppNotFound:
            logger.critical(f"Can not find the Flask application '{build_config.flask_app}' in the file '{build_config.entry_point}'.") # pylint: disable=line-too-long
            logger.critical('Build failed.')
            return False, build_config

        logger.info('Successfully built.')
        return True, config

    def deploy(self, config: Dict[str, str]) -> Tuple[bool, Dict[str, str], WebAppTier]:
        """
        Deploy an application.
        """
        config = self.config_loader(DeploymentConfig.items, config)
        deployment_config = DeploymentConfig(**config)

        tier = self._get_tier(config, deployment_config.tier)

        logger.info('Deploying...')
        self._web_app = _deploy(self.azure_service, deployment_config)
        logger.info('Successfully deployed.')

        return True, config, tier

    def url(self) -> str:
        """
        Get the URL of the deployed application.
        """
        return f'https://{self._web_app.domain}'

    def delete(self, config: Dict[str, str] = None) -> Tuple[bool, Dict[str, str]]:
        """
        Delete the application.
        """
        if config is None:
            config = {'subdomain': self._web_app.name}

        config = self.config_loader(DeleteConfig.items, config)
        delete_config = DeleteConfig(**config)

        logger.info('Deleting...')
        _delete(self.azure_service, delete_config)
        self._web_app = None
        logger.info('Successfully deleted.')

        return True, config

    def github(self, config: Dict[str, str]) -> Tuple[Dict[str, str], Optional[GitHubWorkflow]]:
        """
        Create a GitHub Workflow for the application.
        Workflow is None if the application has been deleted.
        """
        config = self.config_loader(GitHubConfig.items, config)
        github_config = GitHubConfig(**config)
        workflow = github(self.azure_service, github_config)
        # The deleted configuration is not one we want to propagate.
        config.pop('delete', None)
        return config, workflow

    def print_logs(self, config: Dict[str, str]):
        """
        Print the application logs.
        This method is blocking and never returns.
        """
        logs_config = LogsConfig(
            **self.config_loader(LogsConfig.items, config)
        )
        _logs(self.azure_service, logs_config)

    def exists(self) -> bool:
        """
        Return True if the application exists, False otherwise.
        """
        return self._web_app.exists()

    def infrastruture_exists(self) -> bool:
        """
        Return True if the application infrastructure exists, False otherwise.
        """
        return self._web_app.resource_group.exists()

    def all(self, config_loader: Callable[[List[ConfigItem]], Dict[str, str]] = None) -> Iterable['WebApp']:
        """
        Return all WebApp created by WebLodge.
        """
        yield from (
            WebApp(config_loader, self.azure_service, web_app)
            for web_app in _all_az_web_app(self.azure_service)
        )

    def tiers(self, config: Dict[str, str]) -> List[WebAppTier]:
        """
        Return all tiers of the application.
        """
        config = self.config_loader(TiersConfig.items, config)
        tier_config = TiersConfig(**config)
        return _tiers(self.azure_service, tier_config)

    def _get_tier(self, config, tier: str) -> WebAppTier:
        """
        Return the tier of the WebApp.
        """
        config = self.config_loader(TiersConfig.items, config)
        tier_config = TiersConfig(**config)
        tiers = _tiers(self.azure_service, tier_config)

        # Match the tier by name.
        tier = next((t for t in tiers if t.name.upper() == tier), None)

        # If not found, the tier is invalid.
        if not tier:
            raise InvalidTier(
                f"Can not find the tier '{tier}' in the location '{tier_config.location}."
            )

        return tier



"""
Azure Web App representation.
"""
import json
import os
import tempfile
from .interfaces import MicrosoftEntra, MicrosoftEntraApplication
from .resource_group import ResourceGroup
from .cli import Cli


class EntraApplication(MicrosoftEntraApplication):
    """
    Azure Entra Application with Federated Identity.
    """
    def __init__(self, client_id: str, tenant_id: str, subscription_id: str, cli: Cli) -> None:
        super().__init__()
        self.client_id = client_id
        self.tenant_id = tenant_id
        self.subscription_id = subscription_id

        self._cli = cli

    def delete(self):
        """
        Delete the application.
        """
        self._cli.invoke(f'ad app delete --id {self.client_id}', to_json=False)


class Entra(MicrosoftEntra):
    """
    Azure Entra facade.
    """
    _cli = None

    @classmethod
    def set_cli(cls, cli: Cli):
        """
        Set the Azure CLI.
        """
        cls._cli = cli

    # pylint: disable=too-many-arguments
    @classmethod
    def github_application(
        cls,
        name: str,
        username: str,
        repository: str,
        branch: str,
        resource_group: ResourceGroup
    ) -> EntraApplication:
        """
        Create an GitHub Application on Microsoft Entra.

        :param name: The name of the GitHub Application.
        :param branch: The branch of the repository that will trigger the GitHub Action.
        :param username: The username/organisation of the repository owner.
        :param repository: The name of the GitHub repository.
        :param resource_group: The resource group where the application will be deployed.
        :return: The Microsoft Entra representation.
        """
        name = f'weblodge-{name}'

        # Retrieve need informations.
        account = cls._cli.invoke('account show')
        subscription_id = account['id']
        tenant_id = account['tenantId']

        app_id = cls._get_app_id(name)
        service_principal_id = cls._get_sp_id(name, app_id)
        cls._assign_role(subscription_id, service_principal_id, resource_group)
        cls._create_federated_credential(name, app_id, username, repository, branch)

        return EntraApplication(
            client_id=app_id,
            tenant_id=tenant_id,
            subscription_id=subscription_id,
            cli=cls._cli
        )

    @classmethod
    def _get_app_id(cls, name: str) -> str:
        """
        Retrieve the application ID if the application exists, otherwise create it.
        """
        applications = cls._cli.invoke(f'ad app list --display-name {name}')

        for app in applications:
            if app['displayName'] == name:
                return app['appId']

        return cls._cli.invoke(f'ad app create --display-name {name}')['appId']

    @classmethod
    def _get_sp_id(cls, app_name: str, app_id: str) -> str:
        """
        Retrieve the Service Principal of an Application or create it.
        """
        service_principals = cls._cli.invoke(f'ad sp list --display-name {app_name}')

        for service_principal in service_principals:
            if service_principal['appId'] == app_id:
                return service_principal['id']

        return cls._cli.invoke(f'ad sp create --id {app_id}')['id']

    @classmethod
    def _assign_role(cls, subscription_id, service_principal_id, resource_group):
        """
        Assign the owner role to the service principal on the resource group
        if it not already assigned.
        The owner role is required to assign authorizations to KeyVault consumers.
        """
        role_assignments = cls._cli.invoke(
            ' '.join((
                'role assignment list',
                f'--scope {resource_group.id_}',
                f'--assignee {service_principal_id}',
                '--role owner'
            ))
        )
        if role_assignments:
            return

        cls._cli.invoke(
            ' '.join((
                'role assignment create',
                '--role owner',
                f'--subscription {subscription_id}',
                f'--assignee-object-id {service_principal_id}',
                '--assignee-principal-type ServicePrincipal',
                f'--scope {resource_group.id_}'
            )),
            to_json=False
        )

    @classmethod
    def _create_federated_credential(cls, name: str, app_id: str, username: str, repository: str, branch: str):
        """
        Create the federated credential for a GitHub access.
        """
        cred_specs = {
            "name": name,
            "issuer": "https://token.actions.githubusercontent.com",
            "subject": f"repo:{username}/{repository}:ref:refs/heads/{branch}",
            "description": f'WebLodge GitHub Application for application: {name}',
            "audiences": [
                "api://AzureADTokenExchange"
            ]
        }

        creds = cls._cli.invoke(f'ad app federated-credential list --id {app_id}')
        for cred in creds:
            if  cred.get('subject') == cred_specs['subject'] and \
                cred.get('issuer') == cred_specs['issuer'] and \
                cred.get('audiences') == cred_specs['audiences']:
                return

        # pylint: disable=consider-using-with
        credential_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
        json.dump(cred_specs, credential_file)
        credential_file.close()

        cls._cli.invoke(
            ' '.join((
                'ad app federated-credential create',
                f'--id {app_id}',
                f'--parameters {credential_file.name}'
            )),
            to_json=False
        )

        os.unlink(credential_file.name)



"""
GitHub Application on Microsoft Entra.
It allows to deploy an application on Azure WebApplication using GitHub Actions.
"""
from dataclasses import dataclass
from typing import Optional
from weblodge.config import Item as ConfigItem
from weblodge._azure import AzureService

from .build import BuildConfig
from .deploy import DeploymentConfig


@dataclass
class GitHubWorkflow:
    """
    GitHub workflow.
    """
    # The branch that triggers the workflow.
    branch: str
    # Workflow content.
    content: str
    # Application credentials to put in GitHub secrets.
    client_id: str
    tenant_id: str
    subscription_id: str


class GitHubConfig:
    """
    GitHub Application configuration.
    Handle deployment and build configuration.
    """
    # Configurable items of the build.
    items = [
        ConfigItem(
            name='username',
            description='The GitHub username/organisation.',
        ),
        ConfigItem(
            name='repository',
            description='The GitHub repository.',
        ),
        ConfigItem(
            name='branch',
            description='The deployment branch.',
        ),
        ConfigItem(
            name='delete',
            description='Delete the deployment application.',
            attending_value=False,
        ),
        *BuildConfig.items,
        *DeploymentConfig.items
    ]

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        subdomain: str,
        branch: str,
        username: str,
        repository: str,
        location: str,
        delete: bool,
        *_args,
        **_kwargs
    ):
        self.subdomain = subdomain
        self.branch = branch
        self.username = username
        self.repository = repository
        self.location = location
        self.delete = delete


def github(service: AzureService, config: GitHubConfig) -> Optional[GitHubWorkflow]:
    """
    Create a GitHub Application on Microsoft Entra.

    :param service: The Azure service.
    :param config: The GitHub Application configuration.
    :return: The GitHub workflow or None if the GitHub workflow has been deleted.
    """
    rg_created = False
    resource_group = service.resource_groups(config.subdomain)

    if not resource_group.exists():
        rg_created = True
        resource_group.create(config.location)

    entra_application = service.entra.github_application(
        name=config.subdomain,
        branch=config.branch,
        username=config.username,
        repository=config.repository,
        resource_group=resource_group
    )

    if config.delete:
        entra_application.delete()
        if rg_created:
            # The resource group may have been deleted before requesting the application's deletion.
            # In this case, we've recreated it to be able to find the application,
            # and we need to delete it again.
            resource_group.delete()
        return None

    return GitHubWorkflow(
        branch=config.branch,
        content=_github_workflow(config),
        client_id=entra_application.client_id,
        tenant_id=entra_application.tenant_id,
        subscription_id=entra_application.subscription_id,
    )


def _github_workflow(config: GitHubConfig) -> str:
    """
    GitHub workflow.
    """
    return f"""\
# This file is generated by weblodge.
# Please do not edit it manually.
name: Weblodge deploy.
on:
  push:
    branches:
      - {config.branch}

permissions:
  id-token: write
  contents: read

jobs:
  build-and-deploy:
    timeout-minutes: 20
    runs-on: ubuntu-latest
    steps:
      - name: 'Az CLI login'
        uses: azure/login@v1
        with:
          client-id: ${{{{ secrets.AZURE_CLIENT_ID }}}}
          tenant-id: ${{{{ secrets.AZURE_TENANT_ID }}}}
          subscription-id: ${{{{ secrets.AZURE_SUBSCRIPTION_ID }}}}

      - uses: actions/checkout@v4

      - name: 'Install WebLodge.'
        run: |
          pip install weblodge

      - name: 'Build and deploy.'
        run: |
          weblodge deploy --build
"""



"""
A configuration item represents information about a function that can be
overwritten at runtime.
"""
from dataclasses import dataclass


@dataclass(frozen=True)
class Item:
    """
    Information that can be overwritten at runtime.
    """
    # Name of the field.
    # This name will be used as the command line argument, in the code and config files.
    name: str
    # Description of the field. What is it used for?
    description: str
    # Does not expect a value from the user. The argument is a flag.
    # Example: --verbose
    attending_value: str = True
    # Default value of the field.
    # When specified, the field is optional.
    default: str = None
    # List of possible values.
    values_allowed: list = None

    def __eq__(self, obj) -> bool:
        if isinstance(obj, Item):
            return self.name == obj.name
        if isinstance(obj, str):
            return self.name == obj
        return False



"""
SKU class for Azure resources.
"""
from dataclasses import dataclass
from typing import Iterable, Union

from urllib3 import Retry as urllib_retry, request as urllib_request

from .interfaces import AzureAppServiceSku
from .exceptions import InvalidSku, InvalidLocation


# Function to use for HTTP calls and mocks.
RETRY = urllib_retry
REQUEST = urllib_request

# Tier description of SKU families.
_F_TIER = 'Free tier - limited to 60 minutes per day.'
_B_TIER = 'Designed for apps with lower traffic requirements and not needing advanced auto scale and traffic management features.'  # pylint: disable=line-too-long
_S_TIER = 'Designed for running production workloads.'
_PV3_TIER = 'Designed to provide enhanced performance for production apps and workload.'

# Hard coded hardware capabilities for each SKU.
_SKU_INFOS = {
    'F1': {'cores': 1, 'ram': 1, 'disk': 1, 'description': _F_TIER},

    'B1': {'cores': 1, 'ram': 1.75, 'disk': 10, 'description': _B_TIER},
    'B2': {'cores': 2, 'ram': 3.50, 'disk': 10, 'description': _B_TIER},
    'B3': {'cores': 4, 'ram': 7, 'disk': 10, 'description': _B_TIER},

    'S1': {'cores': 1, 'ram': 1.75, 'disk': 50, 'description': _S_TIER},
    'S2': {'cores': 2, 'ram': 3.50, 'disk': 50, 'description': _S_TIER},
    'S3': {'cores': 4, 'ram': 7, 'disk': 50, 'description': _S_TIER},

    'P0v3': {'cores': 1, 'ram': 4, 'disk': 250, 'description': _PV3_TIER},
    'P1v3': {'cores': 2, 'ram': 8, 'disk': 250, 'description': _PV3_TIER},
    'P1mv3': {'cores': 2, 'ram': 16, 'disk': 250, 'description': _PV3_TIER},
    'P2v3': {'cores': 4, 'ram': 16, 'disk': 250, 'description': _PV3_TIER},
    'P2mv3': {'cores': 4, 'ram': 32, 'disk': 250, 'description': _PV3_TIER},
    'P3v3': {'cores': 8, 'ram': 32, 'disk': 250, 'description': _PV3_TIER},
    'P3mv3': {'cores': 8, 'ram': 64, 'disk': 250, 'description': _PV3_TIER},
    'P4mv3': {'cores': 16, 'ram': 128, 'disk': 250, 'description': _PV3_TIER},
    'P5mv3': {'cores': 32, 'ram': 256, 'disk': 250, 'description': _PV3_TIER},
}

# List of available SKU names.
AVAILABLE_SKUS = list(_SKU_INFOS.keys())


@dataclass(frozen=True)
class AppServiceSku(AzureAppServiceSku):
    """
    User representation of the SKU.
    """
    # Name of the SKU.
    name: str

    # Name of the location where the SKU is available.
    location: str

    # Price per hour of the SKU.
    price_by_hour: float

    # Description of the SKU.
    description: str

    # Number of Cores.
    cores: int

    # RAM in GB.
    ram: int

    # Disk size in GB.
    disk: int


def get_skus(location: str) -> Iterable[AzureAppServiceSku]:
    """
    Return availables SKUs for the given location.
    """
    try:
        # Retrieve SKUs from the Azure API.
        skus = REQUEST(
            'GET',
            f"https://prices.azure.com/api/retail/prices?$filter=serviceName eq 'Azure App Service' and contains(productName, 'Linux') and armRegionName eq '{location}' and unitOfMeasure eq '1 Hour' and type eq 'Consumption' and isPrimaryMeterRegion eq true and currencyCode eq 'USD'",  # pylint: disable=line-too-long
            retries=RETRY(total=10, backoff_factor=5, status=5, status_forcelist=[500, 502, 503, 504])
        )
        items = skus.json()['Items']
    except Exception as exception:  # pylint: disable=bare-except
        raise InvalidSku("""Unable to retrieve the list of SKUs.
Please check your internet connection."""
        ) from exception

    # The 'Items' key is empty when the location does not exist.
    if len(items) == 0:
        raise InvalidLocation(f"Can not find any SKU for the location '{location}'.")

    # Return each SKU as AppServiceSKU if find in the hard coded database.
    for item in items:
        sku_info = _SKU_INFOS.get(item['skuName'])

        if sku_info:
            yield AppServiceSku(
                name=item['skuName'],
                location=item['armRegionName'],
                price_by_hour=item['retailPrice'],
                description=sku_info['description'],
                cores=sku_info['cores'],
                ram=sku_info['ram'],
                disk=sku_info['disk']
            )

    # Return the free SKU.
    yield AppServiceSku(
        name='F1',
        location=location,
        price_by_hour=0.0,
        description=_SKU_INFOS['F1']['description'],
        cores=_SKU_INFOS['F1']['cores'],
        ram=_SKU_INFOS['F1']['ram'],
        disk=_SKU_INFOS['F1']['disk']
    )



"""
Azure Resource Group interface.
"""
from typing import Dict

from .resource import Resource
from .interfaces import AzureResourceGroup


class ResourceGroup(Resource, AzureResourceGroup):
    """
    Azure Resource Group representation.
    """
    _cli_prefix = 'group'

    @property
    def id_(self) -> str:
        """
        Return the Resource Group ID.
        """
        return self._from_az['id']

    @property
    def location(self) -> str:
        """
        Return the Resource Group location.
        """
        return self._from_az['location']

    def create(self, location: str, tags: Dict[str, str] = None) -> 'ResourceGroup':
        """
        Create a new Resource Group and return it.
        """
        # Tags are merged with the internal tags.
        tags = {
            **(tags or {}),
            **self._internal_tags
        }

        self._invoke(
            f'{self._cli_prefix} create --name {self.name} --location {location}',
            tags=tags
        )
        return self

    def delete(self) -> None:
        """
        Delete the resource group.
        """
        self._invoke(f'{self._cli_prefix} delete --name {self.name} --yes', to_json=False)

    @classmethod
    def from_az(cls, name: str, from_az: Dict):
        """
        Create the Resource Group from Azure.
        """
        return cls(name, from_az)

    def _load(self):
        """
        Load the Resource Group from Azure.
        """
        self._from_az.update(
            self._invoke(f'{self._cli_prefix} show --name {self.name}')
        )
        return self



"""
Interface to the Azure CLI.
"""
import json
import logging
from io import StringIO
import time
from typing import Dict, List, Union

from azure.cli.core import get_default_cli

from .exceptions import CLIException


logger = logging.getLogger('weblodge')


class Cli:
    """
    Azure CLI wrapper.
    """
    def __init__(self):
        self._first_invoke = True
        self.cli = get_default_cli()

    # pylint: disable=too-many-arguments
    def invoke(
            self,
            command: str,
            to_json=True,
            tags: Dict[str, str] = None,
            log_outputs: bool = False,
            command_args: List[str] = None
        ) -> Union[str, Dict]:
        """
        Execute an Azure CLI command and return its output.
        If `to_json` is True, the output is converted to a JSON object.
        If `log_outputs` is True, the output is not returned but logged instead.
        `command_args` contains the arguments to add to the command as is without split.
        """
        command_args = command_args or []

        if self._first_invoke:
            self._first_invoke = False

            try:
                return self._invoke(
                    command,
                    to_json=to_json,
                    tags=tags,
                    log_outputs=log_outputs,
                    command_args=command_args
                )
            except Exception as exception:  # pylint: disable=broad-exception-caught
                if 'az login' in str(exception):
                    logger.info('Authentication is needed...')
                    self._invoke('login', to_json=False, tags=None, log_outputs=log_outputs, command_args=[])
                    logger.info('Login successful, retry previous command...')

        return self._invoke(command, to_json=to_json, tags=tags, log_outputs=log_outputs, command_args=command_args)

    # pylint: disable=too-many-arguments
    def _invoke(
        self,
        command: str,
        to_json: bool,
        tags: Dict[str, str],
        log_outputs: bool,
        command_args: List[str]
    ) -> Union[str, Dict]:
        # Convert the string command to a array of arguments.
        cmd = command.split()
        cmd.extend(command_args)
        # Redirect the output to a file if the log output is not asked.
        out_fd = None if log_outputs else StringIO()

        if tags:
            cmd.append('--tags')
            cmd.extend(f'{k}={v}' for k, v in tags.items())

        exception = None
        for i in range(1, 15):
            try:
                # Execute the Azure CLI command.
                if self.cli.invoke(cmd, out_file=out_fd):
                    exception = CLIException(f"Error during execution of the command '{command}'.")  # pylint: disable=broad-exception-raised
                else:
                    exception = None
                    break
            except (SystemExit, Exception) as exception: # pylint: disable=broad-exception-caught
                exception = CLIException(f"Error during execution of the command '{command}'.\nTraceback: {exception}") # pylint: disable=raise-missing-from
            time.sleep(i)

        if exception:
            raise exception

        # No output to return.
        if log_outputs:
            return None

        # Retrieve the output.
        output = out_fd.getvalue()
        out_fd.close()

        # Convert the output to a JSON object if needed.
        if to_json:
            return json.loads(output)

        return output



"""
Create the infrastructure that will host the application.

The infrastructure is composed of:
- Resource Group
- App Service Plan
- Web App

All that infrastructure is created in the same Azure location and in the same Azure
Resource Group.
"""
import os
import random
import string
import logging

from weblodge.config import Item as ConfigItem
from weblodge._azure import AzureService, AzureWebApp, AzureLogLevel

from .shared import WEBAPP_TAGS
from .exceptions import NoMoreFreeApplicationAvailable
from .utils import get_webapp, set_webapp_env_var


logger = logging.getLogger('weblodge')


# pylint: disable=too-many-instance-attributes
class DeploymentConfig:
    """
    Deployment configuration.

    Configuration of the infrastructure that will host the application.
    If not provide, a random name will be generated to allow user to deploy an application
    without providing any input.
    """
    # Zip file that contains the user application code.
    package: str = 'azwebapp.zip'
    # Time wait after updating the environment variable.
    env_update_waiting_time: int = 60

    # Configurable items of the deployment.
    items = [
        ConfigItem(
            name='subdomain',
            description='Unique subdomain of the application within Azure. If not provide, a random subdomain is used.',  # pylint: disable=line-too-long
            default=''.join(random.choice(string.ascii_lowercase) for _ in range(20))
        ),
        ConfigItem(
            name='tier',
            description='The application computational power (https://azure.microsoft.com/en-us/pricing/details/app-service/linux/).',  # pylint: disable=line-too-long
            default='F1'
        ),
        ConfigItem(
            name='location',
            description='The physical application location.',
            default='northeurope'
        ),
        ConfigItem(
            name='environment',
            description='The environment of your application.',
            default='production'
        ),
        ConfigItem(
            name='dist',
            description='Folder containing the application zipped.',
            default='dist'
        ),
        ConfigItem(
            name='env_file',
            description='The file containing the environment variable.',
            default='.env'
        ),
        ConfigItem(
            name='log_level',
            description='The log level of the application infrastructure.',
            default='error',
            values_allowed=['error', 'info', 'verbose', 'warning']
        ),
    ]

    # pylint: disable=too-many-arguments
    def __init__(
            self,
            subdomain,
            tier,
            location,
            environment,
            dist,
            env_file,
            log_level,
            *_args,
            **_kwargs
        ):
        # Application subdomain.
        # This name must be unique across all of Azure WebApplication.
        # It will be used as the URL of the application and for created Azure resources.
        self.subdomain = subdomain
        # Application SKU (https://azure.microsoft.com/en-us/pricing/details/app-service/linux/).
        self.tier = tier.upper()
        # Application location.
        self.location = location
        # Application environment.
        self.environment = environment
        # Dist directory containing the application zipped.
        self.dist = dist
        # File containing environment variables.
        self.env_file = env_file
        # Application log level.
        self.log_level = log_level

        # Infrastructure tags.
        self.tags = {
            'environment': self.environment
        }


def deploy(azure_service: AzureService, config: DeploymentConfig) -> AzureWebApp:
    """
    Deploy the application to Azure and return its URL.
    """
    web_app = get_webapp(azure_service, config.subdomain)

    if not web_app.exists():
        logger.info('The infrastructure is being created...')
        web_app.app_service.set_sku(config.tier)
        if not web_app.app_service.exists():
            if web_app.app_service.is_free:
                # Only one free AppService Plan is allowed per Azure subscription and location.
                # Check if a free AppService Plan already exists.
                if free_web_app := azure_service.app_services.get_existing_free(config.location):
                    logger.info('Stopping the deployment. No infrastructure created.')
                    raise NoMoreFreeApplicationAvailable(free_web_app.name)
            if not web_app.resource_group.exists():
                web_app.resource_group.create(
                    location=config.location,
                    tags={
                        **config.tags,
                        **WEBAPP_TAGS
                    }
                )
            web_app.app_service.create()
        if not web_app.keyvault.exists():
            web_app.keyvault.create()
        web_app.create()
        logger.info('The infrastructure is created.')

    logger.info('Setting the log level...')
    log_level = azure_service.log_levels()
    _set_log_level(config, log_level)
    web_app.set_log_level(log_level)
    logger.info('The log level has been set.')

    set_webapp_env_var(web_app, config.env_file, config.env_update_waiting_time)

    logger.info('Uploading the application...')
    web_app.deploy(os.path.join(config.dist, config.package))
    logger.info('The application has been uploaded.')

    return web_app


def _set_log_level(config: DeploymentConfig, log_level: AzureLogLevel) -> None:
    """
    Set the application log level.
    """
    if config.log_level == 'error':
        log_level.error()
    elif config.log_level == 'info':
        log_level.information()
    elif config.log_level == 'verbose':
        log_level.verbose()
    elif config.log_level == 'warning':
        log_level.warning()



"""
Public interface of the Azure module.
"""
from abc import abstractmethod
from typing import Dict, Iterable, Iterator, Optional


class AzureLogLevel:
    """
    Log levels.
    By default, the log level is set to Warning.
    """
    @abstractmethod
    def __init__(self) -> None:
        """
        Initialize the log level.
        """

    @abstractmethod
    def error(self) -> None:
        """
        Set the log level to error.
        """

    @abstractmethod
    def information(self) -> None:
        """
        Set the log level to information.
        """

    @abstractmethod
    def verbose(self) -> None:
        """
        Set the log level to verbose.
        """

    @abstractmethod
    def warning(self) -> None:
        """
        Set the log level to warning.
        """


class AzureResourceGroup:
    """
    Azure Resource Group.
    """
    # Name of the Resource Group.
    name: str

    # Location of the Resouce Group.
    location: str

    # Tags of the Resource Group.
    tags: Dict[str, str]

    @abstractmethod
    def __init__(self, name: str) -> None:
        """
        Initialize the Resource Group.
        """

    @abstractmethod
    def create(self, location: str, tags: Dict[str, str] = None) -> 'AzureResourceGroup':
        """
        Create a new Resource Group and return it.
        """

    @abstractmethod
    def delete(self) -> None:
        """
        Delete the resource group.
        """

    @classmethod
    @abstractmethod
    def all(cls) -> Iterator['AzureResourceGroup']:
        """
        Return all the WebApps.
        """


class AzureKeyVaultSecret:
    """
    Azure KeyVault Secret representation.
    """
    # Secret name.
    name: str

    # Secret value.
    value: str

    # Secret URI.
    uri: str


class AzureKeyVault:
    """
    Azure KeyVault representation.
    """
    # The KeyVault name.
    name: str

    # It's resource group.
    resource_group: AzureResourceGroup

    # Tags of the KeyVault.
    tags: Dict[str, str]

    @abstractmethod
    def __init__(self, name: str, resource_group: AzureResourceGroup):
        """
        Initialize the KeyVault.
        """

    @abstractmethod
    def create(self) -> 'AzureKeyVault':
        """
        Create the Azure KeyVault.
        """

    @abstractmethod
    def delete(self) -> None:
        """
        Delete the Azure KeyVault.
        """

    @abstractmethod
    def exists(self) -> bool:
        """
        Return True if the Key Vault exists.
        False otherwise.
        """

    @abstractmethod
    def get_all(self) -> Iterable[AzureKeyVaultSecret]:
        """
        Return the KeyVault secrets.
        """

    @abstractmethod
    def set(self, name: str, value: str) -> AzureKeyVaultSecret:
        """
        Create a secret.
        """


class AzureAppServiceSku:
    """
    Human representation of the SKU.
    """
    # Name of the SKU.
    name: str

    # Name of the location where the SKU is available.
    location: str

    # Price per hour of the SKU.
    price_by_hour: float

    # Human description of the SKU.
    description: str

    # Number of Cores.
    cores: int

    # RAM in GB.
    ram: int

    # Disk size in GB.
    disk: int


class AzureAppService:
    """
    Azure AppService Plan.
    """
    # Name of the Resource Group.
    name: str

    # Location of the Resouce Group.
    location: str

    # Tags of the Resource Group.
    tags: Dict[str, str]

    # True if the AppService Plan is Free.
    is_free: bool

    # True if the AppService Plan support AlwaysOn.
    always_on_supported: bool


    @abstractmethod
    def __init__(self, name: str, resource_group: AzureResourceGroup) -> None:
        """
        Initialize the Azure App Service.
        """

    @abstractmethod
    def create(self) -> 'AzureAppService':
        """
        Create a Linux AppService with Python.
        """

    @abstractmethod
    def set_sku(self, sku_name: str) -> 'AzureAppService':
        """
        Set the AppService Plan SKU.
        """

    @classmethod
    @abstractmethod
    def all(cls) -> Iterator['AzureAppService']:
        """
        Return all the AppService.
        """

    @classmethod
    @abstractmethod
    def get_existing_free(cls, location: str) -> Optional['AzureAppService']:
        """
        Return the free existing Azure App Service if exists in that location.
        None otherwise.
        """

    @staticmethod
    @abstractmethod
    def skus(location: str) -> Iterable[AzureAppServiceSku]:
        """
        Return the list of available SKUs for the given location.
        """


class AzureWebApp:
    """
    Azure Web App.
    """
    # Name of the WebApp.
    # It is also its unique subdomain on Azure.
    name: str

    # The WebApp location.
    # Ex: northeurope, westeurope, etc.
    location: str

    # WebApp domain.
    # Ex: weblodge.azurewebsites.net
    domain: str

    # Tags of the Web App.
    tags: Dict[str, str]

    # Azure App Service of the WebApp.
    app_service: AzureAppService

    # Azure Resource Group of the WebApp.
    resource_group: AzureResourceGroup

    # Azure KeyVault of the WebApp.
    keyvault: AzureKeyVault

    @abstractmethod
    def __init__(
        self,
        name: str,
        resource_group: AzureResourceGroup,
        app_service: AzureAppService,
        keyvault: AzureKeyVault
    ) -> None:
        """
        Initialize the Azure Web App.
        """

    @abstractmethod
    def create(self) -> 'AzureWebApp':
        """
        Create the WebApp.
        """

    @abstractmethod
    def exists(self) -> bool:
        """
        Return True if the Web App exists.
        False otherwise.
        """

    @classmethod
    @abstractmethod
    def all(cls) -> Iterator['AzureWebApp']:
        """
        Return all the WebApps.
        """

    @abstractmethod
    def set_log_level(self, log_level: AzureLogLevel) -> None:
        """
        Set the WebApp log level.
        """

    @abstractmethod
    def deploy(self, src: str) -> None:
        """
        Deploy an application zipped.
        """

    @abstractmethod
    def logs(self) -> None:
        """
        Stream WebApp logs.
        This is a blocking operation. User must run CTRL+C to stop the process.
        """

    @abstractmethod
    def update_environment(self, env: Dict) -> None:
        """
        Update the WebApp environment variables.
        """

    @abstractmethod
    def deployment_in_progress(self) -> bool:
        """
        True if the WebApp is deploying.
        """

    @abstractmethod
    def restart(self) -> None:
        """
        Restart the WebApp.
        """


class MicrosoftEntraApplication:
    """
    Azure Application registered on the Entra platform.
    """
    client_id: str
    tenant_id: str
    subscription_id: str


class MicrosoftEntra:
    """
    Allow to create Azure Entra Applications.
    """
    # pylint: disable=too-many-arguments
    @classmethod
    @abstractmethod
    def github_application(
        cls,
        name: str,
        username: str,
        repository: str,
        branch: str,
        resource_group: AzureResourceGroup
    ) -> MicrosoftEntraApplication:
        """
        Return a Azure Entra Application for a GitHub Account.
        Credentials are federated based.

        :param name: The name of the GitHub Application.
        :param branch: The branch of the repository that will trigger the GitHub Action.
        :param username: The username/organisation of the repository owner.
        :param repository: The name of the GitHub repository.
        :param resource_group: The resource group where the application will be deployed. Must exists.
        :return: The Microsoft Entra representation.
        """


class AzureService:
    """
    Allow to instanciate Azure components.
    """
    resource_groups: AzureResourceGroup
    app_services: AzureAppService
    web_apps: AzureWebApp
    log_levels: AzureLogLevel
    keyvaults: AzureKeyVault
    entra: MicrosoftEntra



"""
Delete all resources associated with the application.
"""
from weblodge._azure import AzureService
from weblodge.config import Item as ConfigItem


class DeleteConfig:
    """
    Delete configuration.

    Azure Web App names are unique across the entire Azure platform. Therefore, simply providing
    the name is enough to delete the application and all associated resources.
    """
    items = [
        ConfigItem(
            name='subdomain',
            description='The application name to delete.'
        )
    ]

    def __init__(self, subdomain: str, *_args, **_kwargs) -> None:
        self.subdomain = subdomain


def delete(azure_service: AzureService, config: DeleteConfig) -> None:
    """
    Delete the application and corresponding resources.
    """
    azure_service.resource_groups(config.subdomain).delete()



"""
Command Line Interface arguments.
"""
import sys
import argparse
from typing import Tuple

# Default configuration filename.
DEFAULT_CONFIG_FILE = '.weblodge.json'

# Command Line Interface name.
CLI_NAME = sys.argv[0]


def get_cli_args() -> Tuple[str, str]:
    """
    Return the action to perform and the configuration filename.
    """
    # If run user run: `python weblodge.py -h` or `python weblodge.py --help`
    # then we want to display the **WebLodge** `help` message.
    # Otherwise, we want to parse the arguments and better scope the help message.
    # Example: `python weblodge.py build -h` must print the `help` message for the `build` action.
    asking_global_help = len(sys.argv) == 2 and '-h' in sys.argv[1]

    _parser = argparse.ArgumentParser(
        description='Deploy a Python Flask-based application to Azure.',
        add_help=asking_global_help
    )
    _parser.add_argument(
        'action',
        type=str,
        help='Action to perform.',
        choices=['build', 'clean', 'deploy', 'delete', 'github', 'list', 'logs', 'app-tiers']
    )
    _parser.add_argument(
        '--config-file',
        type=str,
        help='File containing the deployment configuration.',
        default=DEFAULT_CONFIG_FILE,
        required=False
    )
    args, _ = _parser.parse_known_args()

    return args.action, args.config_file



"""
Stream logs from an Azure Web App in the user console.
"""
from weblodge.config import Item as ConfigItem
from weblodge._azure import AzureService

from .utils import get_webapp


class LogsConfig():
    """
    Logs configuration.

    Azure Web App names are unique across the entire Azure platform. Therefore, simply providing
    the name is enough to retrieve the application logs.
    """
    items = [
        ConfigItem(
            name='subdomain',
            description='The application subdomain.'
        )
    ]

    def __init__(self, subdomain: str, *_args, **_kwargs) -> None:
        self.subdomain = subdomain


def logs(azure_service: AzureService, config: LogsConfig):
    """
    Stream logs from the application.
    This function is blocking and never returns.
    User must run CTRL+C to stop the process.
    """
    get_webapp(azure_service, config.subdomain).logs()



"""
Azure Service for Azure instanciation.
"""
from .interfaces import AzureResourceGroup, AzureWebApp, AzureAppService, \
    AzureService, AzureLogLevel, MicrosoftEntra, AzureKeyVault

from .cli import Cli
from .entra import Entra
from .web_app import WebApp
from .keyvault import KeyVault
from .log_level import LogLevel
from .appservice import AppService
from .resource_group import ResourceGroup


class Service(AzureService):
    """
    Azure Service.
    Allow to instanciate Azure components.
    """
    resource_groups: AzureResourceGroup
    app_services: AzureAppService
    web_apps: AzureWebApp
    log_levels: AzureLogLevel
    keyvaults: AzureKeyVault
    entra: MicrosoftEntra

    def __init__(self):
        self.cli = Cli()

        self.resource_groups: AzureResourceGroup = ResourceGroup
        self.app_services: AzureAppService = AppService
        self.web_apps: AzureWebApp = WebApp
        self.log_levels: AzureLogLevel = LogLevel
        self.keyvaults: AzureKeyVault = KeyVault
        self.entra: MicrosoftEntra = Entra

        self.resource_groups.set_cli(self.cli)
        self.app_services.set_cli(self.cli)
        self.web_apps.set_cli(self.cli)
        self.keyvaults.set_cli(self.cli)
        self.entra.set_cli(self.cli)



"""
Contains shared variables and functions for the web app.
"""

# Default tags that must be present on all resources.
WEBAPP_TAGS = {
    # The resource group is created by weblodge
    # and contains a WebApp managed by it.
    'contains': 'webapp'
}



"""
Abstract representation of an Azure resource.
"""
import time
import logging
from typing import Dict
from abc import abstractmethod
from collections import UserDict

from .cli import Cli
from .exceptions import AzureException, CanLoadResource

logger = logging.getLogger('weblodge')


class _AzDict(UserDict):
    """
    Lazy loader of the Azure resource.
    """
    def __init__(self, load: callable, **kwargs):
        super().__init__(**kwargs)
        self._load = load

    def __getitem__(self, key):
        if not self.data:
            self._load()
        return super().__getitem__(key)


class Resource:
    """
    Abstract representation of an Azure resource.
    """
    # Prefix used with the Azure CLI.
    # Ex:
    # - 'webapp' for Azure WebApp.
    # - 'appservice plan' for Azure AppService Plan.
    _cli = None
    _cli_prefix = None
    _internal_tags = {'managedby': 'weblodge'}

    def __init__(self, name: str, from_az: Dict = None) -> None:
        self.name = name
        self._from_az = _AzDict(load=self.load, **(from_az or {}))

    @property
    def _user_id(self) -> str:
        """
        Return ID of the current user.
        """
        return self._invoke('account show')['user']['name']

    @property
    def tags(self) -> Dict[str, str]:
        """
        Return the resource tags.
        """
        return self._from_az['tags']

    def __eq__(self, other: object) -> bool:
        return other.name == self.name

    def load(self):
        """
        Load the resource from Azure.
        It automatically retries if the command fails.
        """
        return self._retry(
            self._load,
            f"Cannot load resource '{self.name}' from Azure.",
            CanLoadResource
        )

    def exists(self) -> bool:
        """
        Return True if the resource exists.
        Load the resource from Azure if found.
        """
        for resource in self.all():
            if resource == self:
                self._from_az.update(resource._from_az) # pylint: disable=protected-access
                return True
        return False

    @classmethod
    def all(cls):
        """
        Return all resources managed by WebLodge.
        """
        resources = cls._invoke(
            f'{cls._cli_prefix} list'
        )

        # Tags must be present and not None.
        ressources_with_tags = (_r for _r in resources if _r.get('tags'))
        weblodges_resources = (_r for _r in ressources_with_tags if _r['tags'].get('managedby') == 'weblodge')
        yield from (
            cls.from_az(_r['name'], _r) for _r in weblodges_resources
        )

    @classmethod
    def set_cli(cls, cli: Cli):
        """
        Set the Azure CLI.
        """
        cls._cli = cli

    @classmethod
    def _invoke(cls, *args, **kwargs):
        """
        Return the Azure CLI invoke method.
        Can be transform in operator when support of python3.8 is dropped.
        """
        if cls._cli is None:
            cls._cli = Cli()
        return cls._cli.invoke(*args, **kwargs)

    @classmethod
    @abstractmethod
    def from_az(cls, name: str, from_az: Dict):
        """
        Create the resource from Azure.
        """

    @abstractmethod
    def _load(self):
        """
        Load the resource from Azure.
        """

    # pylint: disable=keyword-arg-before-vararg,too-many-arguments
    @classmethod
    def _retry(
            cls,
            fct,
            log_msg: str,
            exception: AzureException,
            retry: int = 10,
            *args,
            **kwargs
        ):
        """
        Try to execute the given function and retry if it fails.
        """
        try:
            return fct(*args, **kwargs)
        except Exception as raised:  # pylint: disable=broad-exception-caught
            if retry > 0:
                time.sleep(5)
                return cls._retry(
                    fct=fct,
                    log_msg=log_msg,
                    exception=exception,
                    retry=retry - 1,
                    *args,
                    **kwargs
                )
            logger.error(log_msg)
            logger.exception(raised)
            raise exception(log_msg) from raised



"""
Log levels set to the Azure WebApp.
"""

from .interfaces import AzureLogLevel


class LogLevel(AzureLogLevel):
    """
    Log levels.
    By default, the log level is set to Warning.
    """
    def __init__(self) -> None:
        self._log_level = 'warning'

    def to_azure(self) -> str:
        """
        Convert in Azure values.
        """
        return self._log_level

    def error(self) -> None:
        """
        Set the log level to error.
        """
        self._log_level = 'error'

    def information(self) -> None:
        """
        Set the log level to information.
        """
        self._log_level = 'information'

    def verbose(self) -> None:
        """
        Set the log level to verbose.
        """
        self._log_level = 'verbose'

    def warning(self) -> None:
        """
        Set the log level to warning.
        """
        self._log_level = 'warning'



"""
Azure Keyvault interface.
"""
from dataclasses import dataclass
from typing import Dict, Iterable

from weblodge._azure.resource_group import ResourceGroup

from .resource import Resource
from .interfaces import AzureKeyVault, AzureKeyVaultSecret
from .exceptions import CLIException, SecretNotFound


@dataclass(frozen=True)
class KeyVaultSecret(AzureKeyVaultSecret):
    """
    Azure KeyVault secret representation.
    """
    # Secret name.
    name: str
    # Secret value.
    value: str
    # Secret URI.
    uri: str


class KeyVault(Resource, AzureKeyVault):
    """
    Azure KeyVault representation.
    It used to store AzureWebApp secrets.
    It is create only if the App Service
    """
    _cli_prefix = 'keyvault'

    def __init__(self, name: str, resource_group: ResourceGroup, from_az: Dict = None) -> None:
        super().__init__(name, from_az)
        self.resource_group = resource_group

    @property
    def id_(self) -> str:
        """
        Return the KeyVault ID.
        """
        return self._from_az['id']

    def create(self) -> 'AzureKeyVault':
        """
        Create the Azure KeyVault and set the current user as Secret Officer.
        """
        self._from_az = self._invoke(
            ' '.join([
                f'{self._cli_prefix} create',
                f'--location {self.resource_group.location}',
                f'--name {self.name}',
                f'--resource-group {self.resource_group.name}',
                '--enable-rbac-authorization true',
                '--retention-days 7',
            ]),
            tags=self.resource_group.tags
        )
        # Set the current user as Secret Officer.
        self._invoke(
            ' '.join((
                'role assignment create',
                f'--assignee {self._user_id}',
                f'--scope {self.id_}',
            )),
            to_json=False,
            command_args=['--role', 'Key Vault Secrets Officer']
        )
        return self

    def delete(self) -> None:
        """
        Delete the resource group.
        """
        self._invoke(
            f'{self._cli_prefix} delete --name {self.name} --yes',
            to_json=False
        )

    def set(self, name: str, value: str) -> KeyVaultSecret:
        """
        Create or update a secret and return its URI.
        """
        secret = self._invoke(
            ' '.join([
                f'{self._cli_prefix} secret set',
                f'--vault-name {self.name}',
                f'--name {name}'
            ]),
            command_args=['--value', value]
        )
        return KeyVaultSecret(
            uri=secret['id'],
            name=secret['name'],
            value=secret['value']
        )

    def get_all(self) -> Iterable[KeyVaultSecret]:
        """
        Return the KeyVault secrets.
        """
        secrets = self._invoke(f'{self._cli_prefix} secret list --vault-name {self.name}')
        yield from (self._get_secret(s['name']) for s in secrets)

    def can_read_secrets(self, identity: str) -> None:
        """
        Add read permission on the KeyVault secrets.
        """
        self._invoke(
            ' '.join((
                'role assignment create',
                f'--assignee {identity}',
                f'--scope {self.id_}',
            )),
            to_json=False,
            command_args=['--role', 'Key Vault Secrets User']
        )

    @classmethod
    def from_az(cls, name: str, from_az: Dict):
        """
        Create a resource from Azure.
        """
        return cls(
            name=name,
            resource_group=ResourceGroup(from_az['resourceGroup']),
            from_az=from_az
        )

    def _get_secret(self, secret_name: str) -> KeyVaultSecret:
        try:
            secret = self._invoke(
                ' '.join([
                    f'{self._cli_prefix} secret show',
                    f'--vault-name {self.name}',
                    f'--name {secret_name}',
                ])
            )
            return KeyVaultSecret(
                uri=secret['id'],
                name=secret['name'],
                value=secret['value']
            )
        except CLIException as excp:
            raise SecretNotFound(f'Secret "{secret_name}" not found.') from excp

    def _load(self):
        """
        Load the KeyVault from Azure.
        """
        self._from_az.update(
            self._invoke(f'{self._cli_prefix} show --name {self.name}')
        )
        return self



"""
Azure Web App representation.
"""
from typing import Dict

from .resource import Resource
from .log_level import LogLevel
from .appservice import AppService
from .resource_group import ResourceGroup
from .keyvault import KeyVault
from .interfaces import AzureWebApp


class WebApp(Resource, AzureWebApp):
    """
    Azure Web App representation.
    """
    _cli_prefix = 'webapp'

    # pylint: disable=too-many-arguments
    def __init__(
            self,
            name: str,
            resource_group: ResourceGroup,
            app_service: AppService,
            keyvault: KeyVault,
            from_az: Dict = None
        ) -> None:
        super().__init__(name=name, from_az=from_az)
        self.python_version = '3.10'
        self.app_service = app_service
        self.resource_group = resource_group
        self.keyvault = keyvault

    @property
    def location(self) -> str:
        """
        The WebApp location.
        Ex: northeurope, westeurope, etc.
        """
        return self._from_az['location']

    @property
    def domain(self) -> str:
        """
        WebApp domain.
        """
        return self._from_az['hostNames'][0] if self._from_az['hostNames'] else None

    def create(self) -> 'AzureWebApp':
        """
        Create the WebApp infrastructure.

        Settings enabled:
        - WebSockets
        - HTTP/2
        - Always On: If the SKU is not F1.
        - Startup file: weblodge.startup
        """
        name = self.name
        asp = self.app_service.id_
        rg_name = self.resource_group.name
        python_version = self.python_version

        # Create the WebApp infrastructure.
        self._invoke(
            f'{self._cli_prefix} create -g {rg_name} -p {asp} -n {name} --runtime PYTHON:{python_version}',
            tags={**self.resource_group.tags, **self.app_service.tags}
        )
        # Update the WebApp settings.
        self._invoke(
            ' '.join((
                f'{self._cli_prefix} config set --resource-group {rg_name} --name {name}',
                '--web-sockets-enabled true',
                '--http20-enabled',
                '--startup-file weblodge.startup',
                f'--always-on {self.app_service.always_on_supported}',
            ))
        )
        # Retrieve the WebApp identity.
        identity = self._invoke(
            ' '.join((
                f'{self._cli_prefix} identity assign',
                f'-g {self.resource_group.name}',
                f'-n {self.name}',
            ))
        )
        # Allow the WebApp to read the KeyVault secrets.
        self.keyvault.can_read_secrets(identity['principalId'])
        return self

    def set_log_level(self, log_level: LogLevel) -> None:
        """
        Update the log level of the WebApp.
        """
        self._invoke(
            ' '.join((
                f'{self._cli_prefix} log config',
                f'--name {self.name}',
                f'--resource-group {self.resource_group.name}',
                '--application-logging filesystem',
                '--docker-container-logging filesystem',
                '--detailed-error-messages true',
                '--failed-request-tracing true',
                f'--level {log_level.to_azure()}'
            ))
        )

    def deploy(self, src: str) -> None:
        """
        Deploy an application zipped.
        """
        self._invoke(
            ' '.join((
                f'{self._cli_prefix} deployment source config-zip',
                f'-g {self.resource_group.name} -n {self.name}',
                f'--src {src}'
            ))
        )

    def logs(self) -> None:
        """
        Stream WebApp logs.
        This is a blocking operation. User must run CTRL+C to stop the process.
        """
        self._invoke(
            f'{self._cli_prefix} log tail -g {self.resource_group.name} -n {self.name}',
            log_outputs=True
        )

    def update_environment(self, env: Dict) -> None:
        """
        Update the WebApp environment variables.
        """
        env_formatted = []

        # Insert secret in KeyVault.
        for name, value in env.items():
            secret = self.keyvault.set(name, value)
            env_formatted.append(f'{name}=@Microsoft.KeyVault(SecretUri={secret.uri})')

        # Update the WebApp environment variables.
        self._invoke(
            ' '.join((
                f'{self._cli_prefix} config appsettings set',
                f'--name {self.name}',
                f'--resource-group {self.resource_group.name}'
            )),
            to_json=False,
            # Provide as independent arguments to avoid shell escaping issues.
            command_args=['--settings', *env_formatted]
        )

    def deployment_in_progress(self) -> bool:
        """
        True if the WebApp is deploying.
        """
        deployments = self._invoke(
            ' '.join((
                f'{self._cli_prefix} log deployment show',
                f'--name {self.name}',
                f'--resource-group {self.resource_group.name}'
            ))
        )
        return any(
            'Deployment successful' in d.get('message')
            for d in deployments
        )

    def restart(self) -> None:
        """
        Restart the WebApp.
        """
        self._invoke(
            f'{self._cli_prefix} restart -g {self.resource_group.name} -n {self.name}',
            to_json=False
        )

    @classmethod
    def from_az(cls, name: str, from_az: Dict) -> 'AzureWebApp':
        """
        Create a WebApp from Azure result.
        """
        resource_group = ResourceGroup(from_az['resourceGroup'])
        return cls(
            name=name,
            resource_group=resource_group,
            app_service=AppService.from_id(from_az['appServicePlanId']),
            keyvault=KeyVault(name=name, resource_group=resource_group),
            from_az=from_az
        )

    def _load(self):
        """
        Load the WebApp from Azure.
        """
        self._from_az.update(
            self._invoke(
                f'{self._cli_prefix} show --resource-group {self.resource_group.name} --name {self.name}'
            )
        )
        return self



"""
Based on Zip Deploy: https://learn.microsoft.com/en-us/azure/app-service/deploy-zip?tabs=cli

The output generates a zip file containing
- The user application code.
- The user application requirements.
- A generated Kudu deployment configuration file.
- A generated startup file.

This package is ready to be deployed on an Azure Web App.
"""
import os
from pathlib import Path
import zipfile

from weblodge.config import Item as ConfigItem

from .exceptions import RequirementsFileNotFound, EntryPointFileNotFound, FlaskAppNotFound


class BuildConfig:
    """
    Build configuration.

    User-customizable fields are found in the `config` property. All are optional from the user's
    point of view so build can proceed without any configuration.
    """
    # Zip file that contains the user application code.
    package: str = 'azwebapp.zip'
    # Kudu deployment config file.
    kudu_config: str = '.deployment'
    # Startup file.
    # Set in the deployment config too.
    startup_file: str = 'weblodge.startup'
    # Kudu needs a requirements file at the root of the zip.
    kudu_requirements_path = 'requirements.txt'

    # Configurable items of the build.
    items = [
        ConfigItem(
            name='src',
            description='Application folder.',
            default='.'
        ),
        ConfigItem(
            name='dist',
            description='Build destination.',
            default='dist'
        ),
        ConfigItem(
            name='entry_point',
            description='Application entry point.',
            default='app.py'
        ),
        ConfigItem(
            name='flask_app',
            description='The Flask application object.',
            default='app'
        ),
        ConfigItem(
            name='requirements',
            description='Requirements.txt file path.',
            default='requirements.txt'
        )
    ]

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        src: str,
        dist: str,
        entry_point: str,
        flask_app: str,
        requirements: str,
        *_args,
        **_kwargs
    ):
        # Source directory to zip.
        self.src = src
        # Destination directory to the zip file `name`.
        self.dist = dist
        # Application entrypoint.
        self.entry_point = entry_point
        # Flask application object.
        self.flask_app = flask_app
        # User requirements file.
        self.requirements = requirements

    @property
    def package_path(self) -> str:
        """
        Return the package path.
        """
        return os.path.join(self.dist, self.package)


def build(config: BuildConfig) -> None:
    """
    Build an application to a deployable format.
    """
    # Create the destination directory.
    os.makedirs(config.dist, exist_ok=True)

    # Zip all required files together.
    with zipfile.ZipFile(config.package_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
        _user_application(config, zipf)
        _user_requirements(config, zipf)
        _deployment_config(config, zipf)
        _startup_file(config, zipf)


def _user_application(config: BuildConfig, zipf: zipfile.ZipFile):
    """
    Create the zip folder.
    """
    # The requirements file name.
    # It will be added to the zip folder in a dedicated function.
    # If it is added twice, the ZIP library prints an error.
    requirements_filename = Path(config.requirements).name

    # Ensure the entry point exists.
    entry_point = Path(config.src) / config.entry_point
    if not entry_point.exists():
        raise EntryPointFileNotFound()

    # Ensure the flask app is in the entry point file.
    if config.flask_app not in entry_point.read_text():
        raise FlaskAppNotFound()

    for root, _, files in os.walk(config.src):
        root = Path(root)
        # Skip hidden files and directories.
        if root.name.startswith('.'):
            continue
        # Skip bytecode files.
        if '__pycache__' in root.name:
            continue
        # Skip the build directory.
        # Path conversion removes unnecessary slashes.
        if root.name.startswith(Path(config.dist).name):
            continue
        # Zip the files.
        for file in files:
            file_path = root / file
            relative_to = os.path.relpath(file_path, config.src)

            # Skip the requirements file.
            if relative_to == requirements_filename:
                continue

            zipf.write(file_path, relative_to)


def _user_requirements(config: BuildConfig, zipf: zipfile.ZipFile):
    """
    Add the requirements file to the zip folder from the user folder.
    The file can be in the local folder or in the `src` folder.
    """
    if os.path.exists(config.requirements):
        zipf.write(config.requirements, config.kudu_requirements_path)
        return

    requirements_in_src = Path(config.src) / config.requirements
    if requirements_in_src.exists():
        zipf.write(requirements_in_src, config.kudu_requirements_path)
        return

    raise RequirementsFileNotFound()


def _deployment_config(config: BuildConfig, zipf: zipfile.ZipFile):
    """
    Add the deployment config file to the zip folder.
    """
    # Kudu deployment config file.
    kudu_config = '''\
[config]
# Packages must be installed using during the deployment build.
SCM_DO_BUILD_DURING_DEPLOYMENT = true
'''
    # Add the deployment config file to the zip folder.
    zipf.writestr(config.kudu_config, kudu_config)


def _startup_file(config: BuildConfig, zipf: zipfile.ZipFile):
    """
    Add the startup file to the zip folder.
    """
    # Remove potentional .py extension.
    entrypoint = config.entry_point
    if entrypoint.endswith('.py'):
        entrypoint = entrypoint[:-3]

    # Add the application object if it's not already there.
    if ':' not in entrypoint:
        entrypoint = f'{entrypoint}:{config.flask_app}'

    # Default application configuration update with the user and entrypoint.
    # https://learn.microsoft.com/en-us/azure/developer/python/configure-python-web-app-on-app-service
    startup_file_content = f'gunicorn --bind=0.0.0.0 --timeout 600 {entrypoint}'

    # Add the startup file to the zip folder.
    zipf.writestr(config.startup_file, startup_file_content)



"""
Delete all resources created by WebLodge.
"""
from typing import Iterable
from weblodge._azure import AzureService, AzureWebApp

from .utils import get_webapp
from .shared import WEBAPP_TAGS


def _all(azure_service: AzureService) -> Iterable[AzureWebApp]:
    """
    Return all resources with WebApp created by WebLodge.
    In case of error, the WebApp may not be exists, but other resource may exists
    and so the element is returned.
    """
    for resource_group in azure_service.resource_groups.all():
        if WEBAPP_TAGS.items() <= resource_group.tags.items():
            yield get_webapp(azure_service, resource_group.name)



"""
Module exceptions.
"""

class AzureException(Exception):
    """
    Base class for all Azure exceptions.
    """

class CanLoadResource(AzureException):
    """
    Raise when a resource cannot be loaded from Azure.
    """

class ResourceNotFound(AzureException):
    """
    Raise when a resource cannot be found on Azure.
    """

class ResourceGroupNotFound(ResourceNotFound):
    """
    Raise when a resource group cannot be found on Azure.
    """

class AppServiceNotFound(ResourceNotFound):
    """
    Raise when an app service plan cannot be found on Azure.
    """

class CLIException(AzureException):
    """
    Raise when an error occurs while executing an Azure CLI command.
    """

class InvalidSku(AzureException):
    """
    Raise when an invalid SKU is provided.
    """

class InvalidLocation(AzureException):
    """
    Raise when an invalid  is provided.
    """

class SecretNotFound(AzureException):
    """
    Raise when a secret on a KeyVault cannot be found.
    """



"""
Web App exceptions.
"""
class BuildException(Exception):
    """
    Exceptions relative to the build.
    """


class RequirementsFileNotFound(BuildException):
    """
    The requirements file was not found.
    """


class EntryPointFileNotFound(BuildException):
    """
    The entry point file was not found.
    """


class FlaskAppNotFound(BuildException):
    """
    Can not find the Flask application object in the entry point file.
    """


class DeploymentException(Exception):
    """
    Exceptions relative to the deployment.
    """


class NoMoreFreeApplicationAvailable(DeploymentException):
    """
    A free application is already deployed.
    Contains the name of the free application.
    """


class InvalidTier(DeploymentException):
    """
    Can not find the asked tier.
    """


class AppTierException(Exception):
    """
    Exceptions relative to the app-tier command.
    """


class CanNotFindTierLocation(AppTierException):
    """
    Can not find any tier for the provided location.
    """



"""
Azure AppService Plan abstraction.

Allow to CRUD on Azure AppService Plan.
"""
from typing import Dict, Iterable, Optional

from .sku import get_skus, AVAILABLE_SKUS
from .resource import Resource
from .exceptions import InvalidSku
from .resource_group import ResourceGroup
from .interfaces import AzureAppService, AzureAppServiceSku


class AppService(Resource, AzureAppService):
    """
    Azure AppService Plan representation.
    """
    _cli_prefix = 'appservice plan'

    # pylint: disable=too-many-arguments
    def __init__(
            self,
            name: str,
            resource_group: ResourceGroup,
            from_az: Dict = None
        ) -> None:
        super().__init__(name, from_az)
        self.resource_group = resource_group
        self._sku = self._from_az['sku']['name'] if from_az else None

    @property
    def id_(self) -> str:
        """
        Return the AppService Plan ID.
        """
        return self._from_az['id']

    @property
    def always_on_supported(self) -> bool:
        """
        Return True if the AppService Plan support AlwaysOn.
        """
        return not self.is_free

    @property
    def is_free(self) -> bool:
        """
        Return True if the AppService Plan is Free.
        """
        if not self._sku:
            self._sku = self._from_az['sku']['name']
        return self._sku == 'F1'

    @property
    def location(self) -> str:
        """
        Return the AppService Plan location.
        """
        return self.resource_group.location

    def set_sku(self, sku_name: str) -> 'AzureAppService':
        """
        Set the AppService Plan SKU.
        """
        if sku_name not in AVAILABLE_SKUS:
            raise InvalidSku(f"Invalid SKU: '{sku_name}'")

        self._sku = sku_name
        return self

    def create(self) -> 'AzureAppService':
        """
        Create a Linux AppService Plan with Python.
        """
        if not self._sku:
            raise InvalidSku('The SKU must be set before creating the AppService Plan.')

        tags = self.resource_group.tags
        rg_name = self.resource_group.name
        location = self.resource_group.location

        self._from_az = self._invoke(
            f'{self._cli_prefix} create --name {self.name} --sku {self._sku} --resource-group {rg_name} --location {location} --is-linux',  # pylint: disable=line-too-long
            tags=tags
        )
        return self

    @classmethod
    def get_existing_free(cls, location: str) -> Optional['AzureAppService']:
        """
        Return the free existing Azure App Service if exists in that location. None otherwise.
        """
        free_asps = filter(lambda asp: asp.is_free, cls.all())
        with_same_location = filter(lambda asp: asp.location == location, free_asps)
        return next(with_same_location, None)

    @staticmethod
    def skus(location: str) -> Iterable[AzureAppServiceSku]:
        """
        Return the list of available SKUs for the given location.
        """
        yield from get_skus(location)

    @classmethod
    def from_id(cls, id_: str) -> 'AzureAppService':
        """
        Return an App Service from an Azure App Service Plan ID.
        """
        from_az = cls._invoke(f'{cls._cli_prefix} show --ids {id_}')
        return cls.from_az(from_az['name'], from_az)

    @classmethod
    def from_az(cls, name: str, from_az: Dict):
        """
        Return an App Service from Azure AppService result.
        """
        return cls(
            name=name,
            resource_group=ResourceGroup(from_az['resourceGroup']),
            from_az=from_az
        )

    def _load(self):
        """
        Load the AppService Plan from Azure.
        """
        self._from_az.update(
            self._invoke(
                f'{self._cli_prefix} show --name {self.name} --resource-group {self.resource_group.name}'
            )
        )
        self._sku = self._from_az['sku']['name']
        return self
